from spoty import settings
from spoty import log
from spoty.utils import SpotyContext
import spoty.utils
import spoty.spotify_api
import spoty.deezer_api
import spoty.audio_files
import click
import shutil
import os


@click.command("move")
@click.option('--grouping-pattern', '--gp', show_default=True,
              default=settings.SPOTY.DEFAULT_GROUPING_PATTERN,
              help='Tracks will be grouped to playlists according to this pattern (for listing only).')
@click.option('--print-pattern', '--pp', show_default=True,
              default=settings.SPOTY.DEFAULT_PRINT_PATTERN,
              help='Print a list of tracks according to this formatting pattern.')
@click.option('--confirm', '-y', is_flag=True,
              help='Do not ask for confirmation')
@click.option('--path', '--p',
              help='Path where to move audio files.')
@click.pass_obj
def move_tracks(context: SpotyContext,
                grouping_pattern,
                print_pattern,
                confirm,
                path
                ):
    """
Move tracks.
    """
    context.summary.append('Moving:')


    all_tags_list = []
    for i, tags_list in enumerate(context.tags_lists):
        all_tags_list.extend(tags_list)

    if len(all_tags_list) == 0:
        context.summary.append("  No tracks to move.")
    else:
        click.echo('Next tracks will be moved:')
        for i, tags_list in enumerate(context.tags_lists):
            if len(context.tags_lists) > 1:
                click.echo()
                click.echo(
                    f'============================= LIST {i + 1}/{len(context.tags_lists)} =============================')
                click.echo()

            spoty.utils.print_tags_list_grouped(tags_list, print_pattern, grouping_pattern)

        if not confirm:
            click.confirm(f'Are you sure you want to move {len(all_tags_list)} tracks?', abort=True)


        moved_spotify_tracks, moved_deezer_tracks, moved_audio_files, moved_csv_tracks = \
            move_tracks_from_all_sources(all_tags_list, path)

        if len(moved_spotify_tracks) > 0:
            context.summary.append(f'  {len(moved_spotify_tracks)} tracks moved from Spotify.')

        if len(moved_deezer_tracks) > 0:
            context.summary.append(f'  {len(moved_deezer_tracks)} tracks moved from Deezer.')

        if len(moved_audio_files) > 0:
            context.summary.append(f'  {len(moved_audio_files)} audio files moved.')

        if len(moved_csv_tracks) > 0:
            context.summary.append(f'  {len(moved_csv_tracks)} tracks from csv moved.')

    click.echo('\n------------------------------------------------------------')
    click.echo('\n'.join(context.summary))


def move_tracks_from_all_sources(all_tags_list, path):
    spotify_playlists, deezer_playlists, local_audio_files, csv_playlists = spoty.utils.sort_tracks_by_source(
        all_tags_list)

    # todo move deezer, spotify, csv tracks

    moved_audio_files = []
    if len(local_audio_files) > 0:
        for file_name in local_audio_files:
            if spoty.utils.is_valid_file(file_name):
                base_name = os.path.basename(file_name)
                new_file_name = os.path.join(path, base_name)
                if os.path.isfile(new_file_name):
                    new_file_name = spoty.utils.find_empty_file_name(new_file_name)
                os.rename(file_name, new_file_name)
                moved_audio_files.append(file_name)

    moved_spotify_tracks = []
    moved_deezer_tracks = []
    moved_csv_tracks = []

    return moved_spotify_tracks, moved_deezer_tracks, moved_audio_files, moved_csv_tracks



from setuptools import setup, find_packages


def read_requirements():
    with open('requirements.txt') as file:
        content = file.read()
        requirements = content.split('\n')
    return requirements


with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="spoty",
    version="0.1.1",
    description="CLI tool for management of Spotify, Deezer and other music services as well as local music files.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Dmitry Savosh",
    author_email="d.savosh@gmail.com",
    url="https://github.com/dy-sh/spoty",
    license="MIT",
    packages=find_packages(exclude=["spoty.plugins.*", "spoty.plugins","plugins.*", "plugins"]),
    include_package_data=True,
    install_requires=read_requirements(),
    python_requires='>=3.7',
    keywords=["music", "audio", "spotify", "deezer"],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: MIT License",
        "Environment :: Console",
        "Topic :: Multimedia :: Sound/Audio",
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3'
    ],
    entry_points="""
        [console_scripts]
        spoty=spoty.cli:cli
    """
)



from spoty import settings
from spoty import log
import spoty.spotify_api
import spoty.utils
import spoty.csv_playlist
import click
import os
import time
from datetime import datetime


@click.group()
def like():
    r"""Likes management (favorites)."""
    pass


@like.command("count")
def like_count():
    r"""
    Displays the number of liked tracks.

    Examples:

        spoty spotify like count
    """
    count = spoty.spotify_api.get_liked_tracks_count()
    click.echo(f'Liked tracks: {count}')


@like.command("add")
@click.argument("track_ids", nargs=-1)
def like_add(track_ids):
    r"""
    Add tracks to liked tracks.

    TRACK_IDS - list of track IDs or URIs. You can specify one ID or many IDs separated by a space.

    Examples:

        spoty spotify like add  00i9VF7sjSaTqblAuKFBDO

        spoty spotify like add  00i9VF7sjSaTqblAuKFBDO 7cjlfruK9Oqw7k5wAZGO72

        spoty spotify like add https://open.spotify.com/track/00i9VF7sjSaTqblAuKFBDO
    """

    track_ids = list(track_ids)
    spoty.spotify_api.add_tracks_to_liked(track_ids)
    click.echo(f'{len(track_ids)} tracks added to liked')


@like.command("export")
@click.option('--path', '--p', default=settings.SPOTY.DEFAULT_EXPORT_PATH, help='Path to create file')
@click.option('--file-name', '--f', default=settings.SPOTY.DEFAULT_LIKES_EXPORT_FILE_NAME, help='File name to create')
@click.option('--overwrite', '-o', is_flag=True,
              help='Overwrite existing files without asking')
@click.option('--no-timestamp', '-T', is_flag=True,
              help='Do not create a subfolder with the current date and time')
def like_export(path, file_name, overwrite, no_timestamp):
    r"""Export the list of liked tracks to csv file on disk.

    Examples:

        spoty playlist export

        spoty playlist export --file-name "C:\Users\User\Downloads"
    """

    path = os.path.abspath(path)

    if not no_timestamp:
        now = datetime.now()
        date_time_str = now.strftime("%Y_%m_%d-%H_%M_%S")
        path = os.path.join(path, "export-likes-" + date_time_str)

    file_name = os.path.join(path, file_name)

    if os.path.isfile(file_name) and not overwrite:
        if not click.confirm(f'\nFile "{file_name}" already exist. Overwrite?'):
            click.echo("Aborted!")
            log.info(f'Canceled by user (file already exist)')
            return

    liked_tracks = spoty.spotify_api.export_liked_tracks_to_file(file_name)

    click.echo(f'{len(liked_tracks)} liked tracks exported to file: "{file_name}"')

    return file_name


@like.command("import")
@click.argument('file_names', nargs=-1)
@click.option('unlike', '-u', is_flag=True,
              help='Remove imported tracks from liked tracks (invert).')
def like_import(file_names, unlike):
    r"""Import liked tracks to yor library from csv file on disk.

    FILE_NAMES - list of file names to import. You can specify one file or many files separated by a space.

    Examples:

        spoty spotify like import "C:\Users\User\Downloads\export\likes.csv"

        spoty spotify like import "C:\Users\User\Downloads\export\likes1.csv" "C:\Users\User\Downloads\export\likes2.csv"
    """

    if len(file_names) == 0:
        click.echo("Please, specify file names to import.")
        exit()

    all_tracks_in_file = []

    if len(file_names) > 1:
        with click.progressbar(file_names, label='Importing liked tracks') as bar:
            for file_name in bar:
                try:
                    tracks_in_file = spoty.spotify_api.import_likes_from_file(file_name, unlike, False)
                    all_tracks_in_file += tracks_in_file
                except FileNotFoundError:
                    click.echo(f'\nFile does not exist: "{file_name}"')
                except spoty.csv_playlist.CSVFileEmpty as e:
                    log.warning(f'Cant import file "{file_name}". File is empty.')
                    click.echo(f'\nCant import file "{file_name}". File is empty.')
                except spoty.csv_playlist.CSVFileInvalidHeader as e:
                    mess = f'Cant import file "{file_name}". The header of csv table does not contain any of the required ' \
                           f'fields (isrc, spotify_track_id, title).'
                    log.error(mess)
                    click.echo('\n' + mess)
    else:
        for file_name in file_names:
            try:
                tracks_in_file = spoty.spotify_api.import_likes_from_file(file_name, unlike, True)
                all_tracks_in_file += tracks_in_file
            except FileNotFoundError:
                click.echo(f'\nFile does not exist: "{file_name}"')
            except spoty.csv_playlist.CSVFileEmpty as e:
                log.warning(f'Cant import file "{file_name}". File is empty.')
                click.echo(f'\nCant import file "{file_name}". File is empty.')
            except spoty.csv_playlist.CSVFileInvalidHeader as e:
                mess = f'Cant import file "{file_name}". The header of csv table does not contain any of the required ' \
                       f'fields (isrc, spotify_track_id, title).'
                log.error(mess)
                click.echo('\n' + mess)

    click.echo(f'{len(all_tracks_in_file)} liked tracks imported')



from spoty import log
import spoty.utils
import spoty.csv_playlist
import os.path
import click
import time
import re
from spoty import settings, config_path, secrets_file_name
from spotipy.oauth2 import SpotifyOAuth
import spotipy
import datetime

SPOTIFY_CLIENT_ID = settings.default.SPOTIFY_CLIENT_ID
SPOTIFY_CLIENT_SECRET = settings.default.SPOTIFY_CLIENT_SECRET
REDIRECT_URI = settings.SPOTIFY.REDIRECT_URI

sp = None

cache_file_name = os.path.join(config_path, ".cache")


def get_sp():
    if SPOTIFY_CLIENT_ID is None or SPOTIFY_CLIENT_ID == "":
        click.echo(f"SPOTIFY_CLIENT_ID is not set. Please, edit {secrets_file_name}", err=True)
        exit()

    if SPOTIFY_CLIENT_SECRET is None or SPOTIFY_CLIENT_SECRET == "":
        click.echo(f"SPOTIFY_CLIENT_SECRET is not set. Please, edit {secrets_file_name}", err=True)
        exit()

    if REDIRECT_URI is None or REDIRECT_URI == "":
        click.echo(f"REDIRECT_URI is not set. Please, edit {secrets_file_name}", err=True)
        exit()


    global sp
    if sp is None:
        sp = spotipy.Spotify(requests_timeout=30, auth_manager=SpotifyOAuth(
            SPOTIFY_CLIENT_ID,
            SPOTIFY_CLIENT_SECRET,
            REDIRECT_URI,
            scope="user-library-read user-library-modify playlist-modify-private playlist-read-private playlist-modify-public",
            cache_path=cache_file_name
        ))
    return sp


def get_tracks_from_playlists(playlist_ids: list, add_spoty_tags=True):
    all_tracks = []
    all_tags_list = []
    all_received_playlists = []

    requested_playlists = []

    with click.progressbar(playlist_ids, label=f'Reading tracks in {len(playlist_ids)} Spotify playlists') as bar:
        for playlist_id in bar:

            # remove already requested playlists
            if playlist_id in requested_playlists:
                click.echo(f'Spotify playlist {playlist_id} requested twice. In will be skipped.')
                continue

            playlist = get_playlist_with_full_list_of_tracks(playlist_id, add_spoty_tags)
            requested_playlists.append(playlist_id)

            tracks = playlist['tracks']['items']

            tags = read_tags_from_spotify_tracks(tracks)

            all_tracks.extend(tracks)
            all_tags_list.extend(tags)
            all_received_playlists.append(playlist_id)

    return all_tracks, all_tags_list, all_received_playlists


def get_tracks_of_spotify_user(user_id: str, playlists_names_regex: str = None):
    user_id = parse_user_id(user_id)
    if user_id == 'me':
        playlists = get_list_of_playlists()
        click.echo(f'You have {len(playlists)} playlists in Spotify library')
    else:
        playlists = get_list_of_user_playlists(user_id)
        click.echo(f'User {user_id} has {len(playlists)} playlists in Spotify library')

    if playlists_names_regex is not None:
        playlists = list(filter(lambda pl: re.findall(playlists_names_regex, pl['name']), playlists))

    if len(playlists) == 0:
        return [], [], []

    ids = get_playlists_ids(playlists)
    tracks, tags, playlists = get_tracks_from_playlists(ids)

    return tracks, tags, playlists


def find_track_by_isrc(isrc: str, length=None, length_tolerance=settings.SPOTY.COMPARE_LENGTH_TOLERANCE_SEC):
    tracks = find_track_by_query(f'isrc:{isrc}', length, length_tolerance)
    return tracks


def find_track_id_by_isrc(isrc: str, length=None, length_tolerance=settings.SPOTY.COMPARE_LENGTH_TOLERANCE_SEC):
    track = find_track_by_isrc(isrc, length, length_tolerance)
    return track['id'] if track is not None else None


def find_track_by_artist_and_title(artist: str, title: str, length=None,
                                   length_tolerance=settings.SPOTY.COMPARE_LENGTH_TOLERANCE_SEC):
    track = find_track_by_query(f'track:{title} artist:{artist}', length, length_tolerance)
    return track


def find_track_by_query(query: str, length=None, length_tolerance=settings.SPOTY.COMPARE_LENGTH_TOLERANCE_SEC):
    res = get_sp().search(query)
    try:
        tracks = res['tracks']['items']
        if length is not None:
            for track in tracks:
                if abs(int(track['duration_ms']) / 1000 - int(length)) < length_tolerance:
                    return track
            while res['tracks']['next']:
                res = get_sp().next(res['tracks'])
                tracks = res['tracks']['items']
                for track in tracks:
                    if abs(int(track['duration_ms']) / 1000 - int(length)) < length_tolerance:
                        return track
            return None

        return tracks
    except:
        pass
    return None


def find_track_by_id(id: str):
    track = get_sp().track(id)
    return track


def find_track_id_by_artist_and_title(artist: str, title: str, length=None,
                                      length_tolerance=settings.SPOTY.COMPARE_LENGTH_TOLERANCE_SEC):
    track = find_track_by_artist_and_title(artist, title, length, length_tolerance)
    return track['id'] if track is not None else None


def find_missing_track_ids(tags_list: list, ignore_duration=False):
    found = []
    not_found = []

    tracks_without_id = [tags for tags in tags_list if not 'SPOTIFY_TRACK_ID' in tags]

    if len(tracks_without_id) == 0:
        return tags_list, []

    with click.progressbar(tags_list, length=len(tracks_without_id),
                           label=f'Identifying {len(tracks_without_id)} tracks') as bar:
        for tags in tags_list:
            if "SPOTIFY_TRACK_ID" in tags:
                tags['SPOTY_FOUND_BY'] = 'SPOTIFY_TRACK_ID'
                found.append(tags)
                continue

            if "ISRC" in tags:
                if not ignore_duration and 'SPOTY_LENGTH' in tags:
                    id = find_track_id_by_isrc(tags['ISRC'], tags['SPOTY_LENGTH'])
                else:
                    id = find_track_id_by_isrc(tags['ISRC'])
                if id is not None:
                    tags['SPOTIFY_TRACK_ID'] = id
                    tags['SPOTY_FOUND_BY'] = 'ISRC'
                    found.append(tags)
                    bar.update(1)
                    continue

            if "TITLE" in tags and "ARTIST" in tags:
                if not ignore_duration and 'SPOTY_LENGTH' in tags:
                    id = find_track_id_by_artist_and_title(tags['ARTIST'], tags['TITLE'], tags['SPOTY_LENGTH'])
                else:
                    id = find_track_id_by_artist_and_title(tags['ARTIST'], tags['TITLE'])
                if id is not None:
                    tags['SPOTIFY_TRACK_ID'] = id
                    tags['SPOTY_FOUND_BY'] = 'TITLE,ARTIST'
                    found.append(tags)
                    bar.update(1)
                    continue

            not_found.append(tags)
            bar.update(1)

    return found, not_found


def get_playlist(playlist_id: str):
    playlist_id = parse_playlist_id(playlist_id)

    log.info(f'Requesting playlist {playlist_id}')

    try:
        playlist = get_sp().playlist(playlist_id)
        return playlist
    except:
        return None


def remove_invalid_tracks(tracks):
    # remove tracks without id (was deleted from spotify database)
    new_tracks = []
    for track in tracks:
        if track is not None \
                and 'track' in track \
                and track['track'] is not None \
                and 'id' in track['track'] \
                and track['track']['id'] is not None \
                and track['track']['id'] != "":
            new_tracks.append(track)
    return new_tracks


def get_playlist_with_full_list_of_tracks(playlist_id: str, add_spoty_tags=True, show_progressbar=False):
    playlist_id = parse_playlist_id(playlist_id)

    log.info(f'Collecting playlist {playlist_id}')

    playlist = get_sp().playlist(playlist_id)
    total_tracks = playlist["tracks"]["total"]
    tracks = playlist["tracks"]["items"]
    tracks = remove_invalid_tracks(tracks)

    if add_spoty_tags:
        add_spoty_tags_to_tracks([], tracks, playlist_id, playlist['name'])

    log.debug(f'Playlist {playlist_id} have {total_tracks} tracks (playlist name: "{playlist["name"]}")')
    log.debug(f'Collected {len(tracks)}/{total_tracks} tracks in playlist {playlist_id}')

    # if playlist is larger than 100 songs, continue loading it until end
    result = playlist
    if show_progressbar:
        bar = click.progressbar(length=total_tracks,
                                label=f'Reading {total_tracks} tracks from playlist ' + playlist_id)

    while len(tracks) < total_tracks:
        if 'next' in result:
            result = get_sp().next(result)
        else:
            result = get_sp().next(result["tracks"])
        if result is None:
            break
        result['items'] = remove_invalid_tracks(result['items'])
        if add_spoty_tags:
            add_spoty_tags_to_tracks(tracks, result['items'], playlist_id, playlist['name'])
        tracks.extend(result['items'])
        log.debug(f'Collected {len(tracks)}/{total_tracks} tracks in playlist {playlist_id}')
        if show_progressbar:
            bar.update(len(tracks) - bar.pos)

    if show_progressbar:
        bar.finish()
        click.echo()

    playlist["tracks"]["items"] = tracks

    return playlist


def delete_playlist(playlist_id: str, confirm=False):
    playlist_id = parse_playlist_id(playlist_id)
    playlist = get_playlist(playlist_id)
    if playlist is None:
        return False

    if not confirm:
        if not click.confirm(f'Do you want to delete playlist "{playlist["name"]}" ({playlist["id"]})'):
            click.echo("\nCanceled")
            return False
        click.echo()  # for new line

    get_sp().current_user_unfollow_playlist(playlist_id)

    return True


def get_current_user_id():
    return get_sp().me()['id']


def get_list_of_playlists(owned_by_user=True):
    user_id = get_current_user_id()
    playlists = []

    log.info(f'Collecting playlists for current user')

    with click.progressbar(length=100, label='Reading current user spotify playlists') as bar:

        # load first 50 playlists
        result = get_sp().current_user_playlists(limit=50)
        playlists.extend(result['items'])
        total_playlists = result['total']

        log.debug(f'Read {len(playlists)}/{total_playlists} for current user')

        bar.length = total_playlists
        bar.update(len(playlists))

        # load next 50 playlists
        while result['next']:
            result = get_sp().next(result)
            playlists.extend(result['items'])
            bar.update(len(playlists))

            log.debug(f'Read {len(playlists)}/{total_playlists} for current user')

    if owned_by_user:
        playlists = list(filter(lambda pl: pl['owner']['id'] == user_id, playlists))
    else:
        playlists = list(filter(lambda pl: pl['owner']['id'] != user_id, playlists))

    return playlists


def find_playlist_by_name(name: str, only_owned_by_user=True, playlists=None):
    if playlists is None:
        playlists = get_list_of_playlists(only_owned_by_user)
    return list(filter(lambda pl: pl['name'] == name, playlists))


def get_list_of_user_playlists(user_id: str):
    log.info(f'Collecting playlists for user {user_id}')

    with click.progressbar(length=100, label=f'Reading user {user_id} spotify playlists') as bar:
        # load first 50 playlists
        playlists = []
        result = get_sp().user_playlists(user=user_id, limit=50)
        playlists.extend(result['items'])
        total_playlists = result['total']

        log.debug(f'Collected {len(playlists)}/{total_playlists} playlists for user {user_id}')

        bar.length = total_playlists
        bar.update(len(playlists))

        # load next 50 playlists
        while result['next']:
            result = get_sp().next(result)
            playlists.extend(result['items'])
            bar.update(len(playlists))

            log.debug(f'Collected {len(playlists)}/{total_playlists} playlists for user {user_id}')

        return playlists


def create_playlist(name: str):
    log.info(f'Creating new playlist')

    user_id = get_current_user_id()
    new_playlist = get_sp().user_playlist_create(user_id, name)
    id = new_playlist['id']

    log.success(f'New playlist created (id: {id}, name: "{name}")')

    return id


def copy_playlist(playlist_id: str):
    playlist_id = parse_playlist_id(playlist_id)

    log.info(f'Creating a copy of playlist {playlist_id}')

    playlist = get_playlist_with_full_list_of_tracks(playlist_id)
    tracks = playlist["tracks"]["items"]

    ids = get_track_ids(tracks)
    new_playlist_id = create_playlist(playlist['name'])
    tracks_added, import_duplicates, already_exist, invalid_ids \
        = add_tracks_to_playlist_by_ids(new_playlist_id, ids, True)

    log.success(f"Playlist {playlist_id} copy completed ({len(tracks_added)} tracks added).")

    return new_playlist_id, tracks_added


def get_tracks_of_playlist(playlist_id: str, add_spoty_tags=True, playlist_name: str = None):
    playlist_id = parse_playlist_id(playlist_id)

    log.info(f'Collecting tracks from playlist {playlist_id}')

    # load the first 100 songs
    tracks = []
    result = get_sp().playlist_items(playlist_id, additional_types=['track'], limit=100)

    new_tracks = result['items']
    new_tracks = remove_invalid_tracks(new_tracks)
    if (add_spoty_tags):
        add_spoty_tags_to_tracks(tracks, new_tracks, playlist_id, playlist_name)
    tracks.extend(new_tracks)

    log.debug(f'Collected {len(tracks)}/{result["total"]} tracks')

    # if playlist is larger than 100 songs, continue loading it until end
    while result['next']:
        result = get_sp().next(result)

        new_tracks = result['items']
        new_tracks = remove_invalid_tracks(new_tracks)
        if (add_spoty_tags):
            add_spoty_tags_to_tracks(tracks, new_tracks, playlist_id, playlist_name)
        tracks.extend(new_tracks)

        log.debug(f'Collected {len(tracks)}/{result["total"]} tracks')

    return tracks


def add_spoty_tags_to_tracks(tracks: list, new_tracks: list, playlist_id: str, playlist_name: str):
    counter = len(tracks)
    for track in new_tracks:
        track['track']['SPOTY_PLAYLIST_ID'] = playlist_id
        track['track']['SPOTY_PLAYLIST_INDEX'] = str(counter + 1)
        track['track']['SPOTY_SOURCE'] = 'SPOTIFY'
        if playlist_name is not None:
            track['track']['SPOTY_PLAYLIST_NAME'] = playlist_name
        try:
            track['track']['SPOTY_TRACK_ID'] = track['track']['id']
        except:
            pass
        counter += 1
    return counter


def add_tracks_to_playlist_by_tags(playlist_id: str, tags_list: list, allow_duplicates=False):
    playlist_id = parse_playlist_id(playlist_id)

    tags_list = tags_list.copy()

    for i in range(len(tags_list)):
        tags_list[i]['SPOTIFY_TRACK_ID'] = parse_track_id(tags_list[i]['SPOTIFY_TRACK_ID'])

    import_duplicates = []

    if not allow_duplicates:
        tags_list, import_duplicates = spoty.utils.remove_duplicated_tags(tags_list, ['SPOTIFY_TRACK_ID'])
        if len(import_duplicates) > 0:
            log.debug(f'{len(import_duplicates)} duplicates found when adding tracks. It will be skipped.')

    log.info(f'Adding {len(tags_list)} tracks to playlist {playlist_id}')

    already_exist = []

    if not allow_duplicates:
        tracks = get_tracks_of_playlist(playlist_id)
        tags_in_playlist = read_tags_from_spotify_tracks(tracks)
        tags_list, already_exist = spoty.utils.remove_exist_tags(tags_in_playlist, tags_list, ['SPOTIFY_TRACK_ID'])
        if len(already_exist) > 0:
            log.debug(f'{len(already_exist)} tracks already exist and will be skipped.')

    track_ids = get_track_ids_from_tags_list(tags_list)

    i = 0
    next_tracks = []
    while i < len(track_ids):
        next_tracks.append(track_ids[i])
        if len(next_tracks) == 100 or i == len(track_ids) - 1:
            get_sp().playlist_add_items(playlist_id, next_tracks)
            log.debug(f'{len(next_tracks)} tracks added to playlist')
            next_tracks = []
        i += 1

    log.success(f'Adding tracks to playlist {playlist_id} complete (tracks added: {len(tags_list)}')

    return tags_list, import_duplicates, already_exist


def add_tracks_to_playlist_by_ids(playlist_id: str, track_ids: list, allow_duplicates=False):
    playlist_id = parse_playlist_id(playlist_id)

    for i in range(len(track_ids)):
        track_ids[i] = parse_track_id(track_ids[i])

    import_duplicates = []

    if not allow_duplicates:
        track_ids, import_duplicates = spoty.utils.remove_duplicates(track_ids)
        if len(import_duplicates) > 0:
            log.debug(f'{len(import_duplicates)} duplicates found when adding tracks. It will be skipped.')

    log.info(f'Adding {len(track_ids)} tracks to playlist {playlist_id}')

    already_exist = []

    if not allow_duplicates:
        tracks = get_tracks_of_playlist(playlist_id)
        ids_in_playlist = get_track_ids(tracks)
        track_ids, already_exist = spoty.utils.remove_exist(ids_in_playlist, track_ids)
        if len(already_exist) > 0:
            log.debug(f'{len(already_exist)} tracks already exist and will be skipped.')

    tracks_added = []

    invalid_ids = []

    i = 0
    next_tracks = []
    while i < len(track_ids):
        next_tracks.append(track_ids[i])
        if len(next_tracks) == 100 or i == len(track_ids) - 1:
            try:
                get_sp().playlist_add_items(playlist_id, next_tracks)
                tracks_added.extend(next_tracks)
            except:
                for track_id in next_tracks:
                    try:
                        get_sp().playlist_add_items(playlist_id, [track_id])
                        tracks_added.append(track_id)
                    except:
                        invalid_ids.append(track_id)
                        click.echo(f'"\nTrack ID "{track_id}" not found in spotify database.')
            log.debug(f'{len(next_tracks)} tracks added to playlist')
            next_tracks = []
        i += 1

    log.success(f'Adding tracks to playlist {playlist_id} complete (tracks added: {len(next_tracks)}')

    return tracks_added, import_duplicates, already_exist, invalid_ids


def remove_all_tracks_from_playlist(playlist_id: str, confirm=False):
    playlist_id = parse_playlist_id(playlist_id)
    tracks = get_tracks_of_playlist(playlist_id)

    if len(tracks) == 0:
        return False

    if not confirm:
        if not click.confirm(f'Do you want to remove all tracks from playlist {playlist_id}?'):
            click.echo("\nCanceled")
            return False
        click.echo()  # for new line

    ids = get_track_ids(tracks)
    remove_tracks_from_playlist(playlist_id, ids)
    return True


def get_missing_tracks_from_playlist(playlist_id: str, track_ids: list, confirm=False):
    playlist_id = parse_playlist_id(playlist_id)
    tracks = get_tracks_of_playlist(playlist_id)

    if len(tracks) == 0:
        return []

    missing_tracks = []
    for track in tracks:
        if track['track']['id'] not in track_ids:
            missing_tracks.append(track)

    return missing_tracks


def remove_tracks_from_playlist(playlist_id: str, track_ids: list):
    playlist_id = parse_playlist_id(playlist_id)

    log.info(f'Removing {len(track_ids)} tracks from playlist {playlist_id}')

    i = 0
    next_tracks = []
    while i < len(track_ids):
        track_ids[i] = parse_track_id(track_ids[i])
        next_tracks.append(track_ids[i])
        if len(next_tracks) == 100 or i == len(track_ids) - 1:
            snapshot = get_sp().playlist_remove_all_occurrences_of_items(playlist_id, next_tracks)
            next_tracks = []
        i += 1

    log.success(f'Tracks removed from playlist {playlist_id}')


def remove_liked_tracks_in_playlist(playlist_id: str):
    playlist_id = parse_playlist_id(playlist_id)

    log.info(f'Removing liked tracks from playlist {playlist_id}')

    tracks = get_tracks_of_playlist(playlist_id)
    ids = get_track_ids(tracks)
    liked_track_ids = get_liked_track_ids(ids)
    remove_tracks_from_playlist(playlist_id, liked_track_ids)

    log.success(f'{len(liked_track_ids)} liked tracks removed from playlist {playlist_id}.')

    return liked_track_ids


def get_invalid_tracks_in_playlist(playlist_id: str):
    playlist_id = parse_playlist_id(playlist_id)

    log.info(f'Listing invalid tracks in playlist {playlist_id}')

    invalid_tracks = []

    tracks = get_tracks_of_playlist(playlist_id)
    for track in tracks:
        try:
            id = track['track']['id']
            assert id is not None, "Track id is None"
        except:
            invalid_tracks.append(track)

    return invalid_tracks


def import_playlists_from_tags_list(tags_list: list, grouping_pattern: str, overwrite_if_exist=False,
                                    append_if_exist=False,
                                    allow_duplicates=True, confirm=False):
    grouped_tags = spoty.utils.group_tags_by_pattern(tags_list, grouping_pattern)
    all_playlist_ids, all_added, all_source_duplicates, all_already_exist = \
        import_playlists_from_tag_groups(grouped_tags, overwrite_if_exist, append_if_exist, allow_duplicates, confirm)

    return all_playlist_ids, all_added, all_source_duplicates, all_already_exist


def import_playlists_from_tag_groups(grouped_tags: dict, overwrite_if_exist=False,
                                     append_if_exist=False,
                                     allow_duplicates=True, confirm=False):
    all_playlist_ids = []
    all_added = []
    all_source_duplicates = []
    all_already_exist = []

    playlists = None
    if append_if_exist or overwrite_if_exist:
        playlists = get_list_of_playlists(True)

    with click.progressbar(grouped_tags.items(), label=f'Importing {len(grouped_tags)} playlists') as bar:
        for group_name, g_tags_list in bar:
            playlist_id, added, source_duplicates, already_exist \
                = import_playlist_from_tags_list(group_name, g_tags_list, overwrite_if_exist, append_if_exist,
                                                 allow_duplicates, confirm, playlists)

            all_playlist_ids.append(playlist_id)
            all_added.extend(added)
            all_source_duplicates.extend(source_duplicates)
            all_already_exist.extend(already_exist)

    return all_playlist_ids, all_added, all_source_duplicates, all_already_exist


def import_playlist_from_tags_list(playlist_name: str, tags_list: list, overwrite_if_exist=False, append_if_exist=False,
                                   allow_duplicates=True, confirm=False, playlists=None):
    if len(tags_list) == 0:
        return [], [], [], []

    log.info(f'Importing playlist "{playlist_name}"')

    playlist_id = None

    if overwrite_if_exist or append_if_exist:
        found_playlists = find_playlist_by_name(playlist_name, True, playlists)

        if len(found_playlists) > 0:
            if append_if_exist:
                if len(found_playlists) > 1:
                    if confirm:
                        click.echo(
                            f'\n{len(found_playlists)} playlists with name "{playlist_name}" found in Spotify library. Choosing the first one.')
                    else:
                        if not click.confirm(
                                f'\n{len(found_playlists)} playlists with name "{playlist_name}" found in Spotify library. Choose the first one?'):
                            click.echo("\nCanceled")
                            log.info(f'Canceled by user (more than one playlists found with name "{playlist_name})"')
                            return [], [], [], []
                        click.echo()  # for new line
            if overwrite_if_exist:
                if len(found_playlists) > 1:
                    if confirm:
                        click.echo(
                            f'\n{len(found_playlists)} playlists with name "{playlist_name}" found in Spotify library. Choosing the first one..')
                    else:
                        if not click.confirm(
                                f'\n{len(found_playlists)} playlists with name "{playlist_name}" found in Spotify library. Choose the first one?'):
                            click.echo("\nCanceled")
                            log.info(f'Canceled by user (more than one playlists found with name "{playlist_name})"')
                            return [], [], [], []
                        click.echo()  # for new line

                new_ids = get_track_ids_from_tags_list(tags_list)
                missing_tracks = get_missing_tracks_from_playlist(found_playlists[0]['id'], new_ids, True)
                if len(missing_tracks) > 0:
                    if not confirm:
                        if not click.confirm(
                                f'\nDo you want to remove {len(missing_tracks)} tracks from playlist "{playlist_name}" (these tracks are not in the provided new track list)?'):
                            click.echo("\nCanceled")
                            return [], [], [], []
                        click.echo()  # for new line

                    ids = get_track_ids(missing_tracks)
                    remove_tracks_from_playlist(found_playlists[0]['id'], ids)

            playlist_id = found_playlists[0]['id']

    if playlist_id is None:
        playlist_id = create_playlist(playlist_name)

    added, source_duplicates, already_exist = add_tracks_to_playlist_by_tags(playlist_id, tags_list, allow_duplicates)

    log.success(f'Playlist imported (new tracks: "{len(added)}")  id: {playlist_id} name: "{playlist_name}"')

    return playlist_id, added, source_duplicates, already_exist


def like_all_tracks_in_playlist(playlist_id: str):
    playlist_id = parse_playlist_id(playlist_id)

    log.info(f'Adding likes to all tracks in playlist {playlist_id}')

    tracks = get_tracks_of_playlist(playlist_id)
    ids = get_track_ids(tracks)
    not_liked_track_ids = get_not_liked_track_ids(ids)
    add_tracks_to_liked(not_liked_track_ids)

    log.success(f'{len(not_liked_track_ids)} tracks added to liked tracks in playlist {playlist_id}.')

    return not_liked_track_ids


def get_liked_tracks_count():
    results = get_sp().current_user_saved_tracks(limit=50)
    total_likes_count = results['total']
    return total_likes_count


def get_liked_tracks():
    tracks = []

    log.info(f'Collecting liked tracks')

    with click.progressbar(length=100, label='Collecting liked tracks') as bar:
        results = get_sp().current_user_saved_tracks(limit=50)
        total_likes_count = results['total']
        tracks.extend(results['items'])

        bar.length = total_likes_count
        bar.update(len(tracks))

        log.debug(f'Collected {len(tracks)}/{total_likes_count}')

        while results['next']:
            results = get_sp().next(results)
            tracks.extend(results['items'])
            bar.update(len(tracks))

            log.debug(f'Collected {len(tracks)}/{total_likes_count}')

    return tracks


def add_tracks_to_liked(track_ids: list, show_progress_bar=False):
    for i in range(len(track_ids)):
        track_ids[i] = parse_track_id(track_ids[i])

    log.info(f'Adding {len(track_ids)} to saved tracks')

    i = 0
    next_tracks = []

    if not show_progress_bar:
        while i < len(track_ids):
            next_tracks.append(track_ids[i])
            if len(next_tracks) == 50 or i == len(track_ids) - 1:
                get_sp().current_user_saved_tracks_add(tracks=next_tracks)
                log.debug(f'Added {i + 1}/{len(next_tracks)} tracks to saved')
                next_tracks = []
            i += 1
    else:
        with click.progressbar(length=len(track_ids), label=f'Adding {len(track_ids)} liked tracks') as bar:
            while i < len(track_ids):
                next_tracks.append(track_ids[i])
                if len(next_tracks) == 50 or i == len(track_ids) - 1:
                    get_sp().current_user_saved_tracks_add(tracks=next_tracks)
                    log.debug(f'Added {i + 1}/{len(next_tracks)} tracks to saved')
                    bar.update(len(next_tracks))
                    next_tracks = []
                i += 1


def remove_tracks_from_liked(track_ids: list, show_progress_bar=False):
    for i in range(len(track_ids)):
        track_ids[i] = parse_track_id(track_ids[i])

    log.info(f'Adding {len(track_ids)} to saved tracks')

    i = 0
    next_tracks = []

    if not show_progress_bar:
        while i < len(track_ids):
            next_tracks.append(track_ids[i])
            if len(next_tracks) == 50 or i == len(track_ids) - 1:
                get_sp().current_user_saved_tracks_delete(tracks=next_tracks)
                log.debug(f'Removed {i + 1}/{len(next_tracks)} tracks from saved')
                next_tracks = []
            i += 1
    else:
        with click.progressbar(length=len(track_ids), label='Adding liked tracks') as bar:
            while i < len(track_ids):
                next_tracks.append(track_ids[i])
                if len(next_tracks) == 50 or i == len(track_ids) - 1:
                    get_sp().current_user_saved_tracks_delete(tracks=next_tracks)
                    log.debug(f'Removed {i + 1}/{len(next_tracks)} tracks from saved')
                    bar.update(len(next_tracks))
                    next_tracks = []
                i += 1


def export_liked_tracks_to_file(file_name: str):
    log.info(f'Exporting liked tracks from file "{file_name}"')

    liked_tracks = get_liked_tracks()
    tag_tracks = read_tags_from_spotify_tracks(liked_tracks)
    spoty.csv_playlist.write_tags_to_csv(tag_tracks, file_name)

    log.success(f'{len(liked_tracks)} liked tracks exported to file: "{file_name}"')

    return liked_tracks


def import_likes_from_file(file_name: str, invert=False, show_progressbar=False):
    log.info(f'Importing liked tracks from file "{file_name}"')
    tags_list = spoty.csv_playlist.read_tags_from_csv(file_name)
    ids = spoty.spotify_api.get_track_ids_from_tags_list(tags_list)

    # remove already exist
    if not invert:
        ids = get_not_liked_track_ids(ids, show_progressbar)
    else:
        ids = get_liked_track_ids(ids, show_progressbar)

    if len(ids) > 0:
        if not invert:
            add_tracks_to_liked(ids, show_progressbar)
        else:
            remove_tracks_from_liked(ids, show_progressbar)

    log.success(f'{len(tags_list)} liked tracks imported from file: "{file_name}"')

    return tags_list


def get_likes_for_tracks(track_ids: list, show_progressbar=False):
    likes = []

    i = 0
    next_tracks = []

    if show_progressbar:
        bar = click.progressbar(length=len(track_ids), label='Collecting liked tracks')

    while i < len(track_ids):
        next_tracks.append(track_ids[i])
        if len(next_tracks) == 50 or i == len(track_ids) - 1:
            likes_new = get_sp().current_user_saved_tracks_contains(tracks=next_tracks)
            likes.extend(likes_new)
            if show_progressbar:
                bar.update(len(next_tracks))
            next_tracks = []
        i += 1

    if show_progressbar:
        bar.finish()
        click.echo()

    return likes


def get_liked_track_ids(track_ids: list, show_progressbar=False):
    liked_tracks = []

    likes = get_likes_for_tracks(track_ids, show_progressbar)
    for i in range(len(track_ids)):
        if likes[i]:
            liked_tracks.append(track_ids[i])

    return liked_tracks


def get_not_liked_track_ids(track_ids: list, show_progressbar=False):
    not_liked_tracks = []

    likes = get_likes_for_tracks(track_ids, show_progressbar)
    for i in range(len(track_ids)):
        if not likes[i]:
            not_liked_tracks.append(track_ids[i])

    return not_liked_tracks


def get_liked_tags_list(new_sub_tags_list, show_progressbar=False):
    liked_tags_list = []
    not_liked_tags_list = []
    track_ids = get_track_ids_from_tags_list(new_sub_tags_list)
    likes = get_likes_for_tracks(track_ids, show_progressbar)
    for i in range(len(track_ids)):
        if likes[i]:
            liked_tags_list.append(new_sub_tags_list[i])
        else:
            not_liked_tags_list.append(new_sub_tags_list[i])

    return liked_tags_list, not_liked_tags_list


def get_track_artist_and_title(track: dict):
    artists = list(map(lambda artist: artist['name'], track['artists']))
    artists_str = ', '.join(artists)
    artists_str.replace(' - ', ' ')
    title = track['name'].replace(' - ', ' ')
    return f"{artists_str} - {title}"


def get_track_ids(tracks: list):
    if len(tracks) == 0:
        return []
    else:
        return [item['track']['id'] for item in tracks]


def get_track_ids_from_tags_list(tags_list):
    if len(tags_list) == 0:
        return []
    else:
        return [str(track['SPOTIFY_TRACK_ID']) for track in tags_list]


def get_playlists_ids(playlists: list):
    if len(playlists) == 0:
        return []
    else:
        return [item['id'] for item in playlists]


def parse_playlist_id(id_or_uri: str):
    if id_or_uri.startswith("https://open.spotify.com/playlist/"):
        id_or_uri = id_or_uri.split('/playlist/')[1]
        id_or_uri = id_or_uri.split('?')[0]
    return id_or_uri


def parse_track_id(id_or_uri: str):
    if id_or_uri.startswith("https://open.spotify.com/track/"):
        id_or_uri = id_or_uri.split('/track/')[1]
        id_or_uri = id_or_uri.split('?')[0]
    return id_or_uri


def parse_user_id(id_or_uri: str):
    if id_or_uri.startswith("https://open.spotify.com/user/"):
        id_or_uri = id_or_uri.split('/user/')[1]
        id_or_uri = id_or_uri.split('?')[0]
    return id_or_uri


def read_tags_from_spotify_tracks(tracks: list):
    tag_tracks = []

    for track in tracks:
        tags = read_tags_from_spotify_track(track)
        tag_tracks.append(tags)

    return tag_tracks


def read_tags_from_spotify_track(track: dict):
    date_added = track['added_at'] if "added_at" in track else None

    if "track" in track:
        track = track['track']

    tags = {}

    try:
        tags['ISRC'] = track['external_ids']['isrc']
    except:
        pass

    try:
        artists = list(map(lambda artist: artist['name'], track['artists']))
        tags['ARTIST'] = ';'.join(artists)
    except:
        pass

    tags['TITLE'] = track['name']

    try:
        tags['ALBUM'] = track['album']['name']
    except:
        pass

    tags['SPOTY_LENGTH'] = str(int(int(track['duration_ms']) / 1000))

    try:
        tags['SPOTIFY_ALBUM_ID'] = track['album']['id']
    except:
        pass

    try:
        tags['WWWAUDIOFILE'] = track['external_urls']['spotify']
    except:
        pass

    tags['SPOTIFY_TRACK_ID'] = track["id"]

    tags['EXPLICIT'] = track['explicit']

    tags['TRACK'] = track['track_number']

    try:
        tags['YEAR'] = track['album']['release_date']
    except:
        pass

    # tags['SPOTIFY_DATE_ADDED'] = date_added
    if date_added is not None:
        timestamp = datetime.datetime.strptime(date_added, "%Y-%m-%dT%H:%M:%SZ")  # format: 2021-12-19T19:35:17Z
        tags['SPOTY_TRACK_ADDED'] = timestamp.strftime('%Y-%m-%d %H:%M:%S')

    for tag in spoty.utils.spoty_tags:
        if tag in track:
            tags[tag] = track[tag]

    tags = spoty.utils.clean_tags_after_read(tags)

    return tags



from spoty import settings
from spoty import log
from spoty.utils import SpotyContext
import spoty.utils
import click


@click.command("print")
@click.option('--print-pattern', '--pp', show_default=True,
              help='Print a list of tracks according to this formatting pattern. If not specified, DUPLICATE_PRINT_PATTERN setting from the config file will be used.')
@click.pass_obj
def print_duplicates(context: SpotyContext,
                     print_pattern,
                     ):
    """
Print a list of duplicates to console.
    """

    for i, group in enumerate(context.duplicates_groups):
        click.echo()
        click.echo(f"--------------------------- GROUP {i+1}/{len(context.duplicates_groups)} ---------------------------")
        if len(group.source_tags.items())>0:
            click.echo("Source:")
            spoty.utils.print_duplicates_tags_list([group.source_tags], print_pattern)

        if len(group.def_duplicates) > 0:
            click.echo("Duplicates:")
            spoty.utils.print_duplicates_tags_list(group.def_duplicates, print_pattern)

        if len(group.prob_duplicates) > 0:
            click.echo("Probably duplicates:")
            spoty.utils.print_duplicates_tags_list(group.prob_duplicates, print_pattern)


    click.echo('\n------------------------------------------------------------')
    click.echo('\n'.join(context.summary))



from spoty.commands.duplicates_commands import \
    add_missing_tags_command, \
    export_duplicates_command, \
    print_duplicates_command, \
    get_duplicates_group, \
    get_unique_group
from spoty.utils import SpotyContext
from spoty import settings
from spoty import log
import spoty.deezer_api
import spoty.csv_playlist
import spoty.deezer_api
import spoty.spotify_api
import spoty.audio_files
import spoty.utils
import click
import os
from datetime import datetime


@click.group("dup")
@click.option('--compare-tags-def', '--ctd', show_default=True, multiple=True,
              default=settings.SPOTY.COMPARE_TAGS_DEFINITELY_DUPLICATE,
              help='Compare definitely duplicates by this tags. It is optional. You can also change the list of tags in the config file.')
@click.option('--compare-tags-prob', '--ctp', show_default=True, multiple=True,
              default=settings.SPOTY.COMPARE_TAGS_PROBABLY_DUPLICATE,
              help='Compare probably duplicates by this tags. It is optional. You can also change the list of tags in the config file.')
@click.option('--only-def', '--def', '-d', is_flag=True,
              help='Find only definitely duplicates.')
@click.option('--only-prob', '--prob', '-p', is_flag=True,
              help='Find only probably duplicates.')
@click.pass_obj
def find_duplicates(context: SpotyContext,
                    compare_tags_def,
                    compare_tags_prob,
                    only_def,
                    only_prob
                    ):
    """
Find duplicates.
    """

    tags_list = context.tags_lists[0]

    compare_tags_def = spoty.utils.tuple_to_list(compare_tags_def)
    compare_tags_prob = spoty.utils.tuple_to_list(compare_tags_prob)

    if only_def:
        compare_tags_prob = []
    if only_prob:
        compare_tags_def = []

    duplicates_groups, unique_tracks = spoty.utils.find_duplicates_in_tag_list2(tags_list, compare_tags_def,
                                                                                compare_tags_prob, True)

    context.duplicates_groups = duplicates_groups
    context.unique_first_tracks = unique_tracks

    total_def_duplicates_count = 0
    total_prob_duplicates_count = 0

    for group in context.duplicates_groups:
        total_def_duplicates_count += len(group.def_duplicates)
        total_prob_duplicates_count += len(group.prob_duplicates)

    context.summary.append("Finding duplicates:")
    if total_def_duplicates_count > 0:
        context.summary.append(f'  {total_def_duplicates_count} definitely duplicates found')
    if total_prob_duplicates_count > 0:
        context.summary.append(f'  {total_prob_duplicates_count} probably duplicates found')


find_duplicates.add_command(add_missing_tags_command.add_missing_tags)
find_duplicates.add_command(export_duplicates_command.export_duplicates)
find_duplicates.add_command(print_duplicates_command.print_duplicates)
find_duplicates.add_command(get_duplicates_group.get_duplicates)
find_duplicates.add_command(get_unique_group.get_unique)


from spoty import log
import spoty.utils
import spoty.audio_files
import click
import os
import time
import re


def is_m3u8(file_name):
    return file_name.upper().endswith('.M3U8')


def create_m3u8s(tags_list, path, grouping_pattern, overwrite=False, append=False, remove_duplicates=False,
                 confirm=False, compare_tags_list=None):
    path = os.path.abspath(path)

    if compare_tags_list is None:
        compare_tags_list = []

    all_created_m3u8_file_names = []
    all_created_m3u8_names = []
    all_added_tags = []
    all_import_duplicates = []
    all_already_exist = []

    if len(tags_list) > 0:
        grouped_tags = spoty.utils.group_tags_by_pattern(tags_list, grouping_pattern)

        for group, tags_l in grouped_tags.items():
            m3u8_name = spoty.utils.slugify_file_pah(group)
            m3u8_file_name = os.path.join(path, m3u8_name + '.m3u8')

            create_new_file = True
            if os.path.isfile(m3u8_file_name):
                if append:
                    create_new_file = False
                elif overwrite:
                    if not confirm:
                        if not click.confirm(f'File "{m3u8_file_name}" already exist. Overwrite?'):
                            click.echo("\nCanceled")
                            continue
                        click.echo("")  # for new line
                else:
                    m3u8_file_name = spoty.utils.find_empty_file_name(m3u8_file_name)

            if remove_duplicates:
                for compare_tags_str in compare_tags_list:
                    compare_tags = compare_tags_str.split(',')
                    tags_l, import_duplicates = spoty.utils.remove_duplicated_tags(tags_l, compare_tags, False)
                    all_import_duplicates.extend(import_duplicates)

                if len(all_import_duplicates) > 0:
                    log.debug(f'{len(all_import_duplicates)} duplicates found when adding tracks. It will be skipped.')

                if not create_new_file:
                    exist_tags_list = read_tags_from_m3u8(m3u8_file_name)
                    for compare_tags_str in compare_tags_list:
                        compare_tags = compare_tags_str.split(',')
                        tags_l, exist_tags = spoty.utils.remove_exist_tags(exist_tags_list, tags_l, compare_tags, False)
                        all_already_exist.extend(exist_tags)

            write_tags_to_m3u8(tags_l, m3u8_file_name, not create_new_file)

            all_created_m3u8_names.append(m3u8_name)
            all_created_m3u8_file_names.append(m3u8_file_name)
            all_added_tags.extend(tags_l)

    return all_created_m3u8_file_names, all_created_m3u8_names, all_added_tags, all_import_duplicates, all_already_exist


def find_m3u8s_in_paths(paths, recursive=True):
    all_m3u8s = []

    for path in paths:
        file_names = find_m3u8s_in_path(path, recursive)
        all_m3u8s.extend(file_names)

    return all_m3u8s


def find_m3u8s_in_path(path, recursive=True):
    path = os.path.abspath(path)

    if not spoty.utils.is_valid_path(path):
        raise Exception("Path is not valid: " + path)

    full_file_names = []
    if recursive:
        full_file_names = [os.path.join(dp, f) for dp, dn, filenames in os.walk(path) for f in filenames if
                           is_m3u8(os.path.splitext(f)[1])]
    else:
        full_file_names = [os.path.join(path, f) for f in os.listdir(path) if os.path.isfile(os.path.join(path, f))]
        full_file_names = list(filter(lambda f: is_m3u8(f), full_file_names))

    return full_file_names


def write_tags_to_m3u8(tags_list, m3u8_file_name, append=False):
    tags_list = spoty.utils.clean_tags_list_before_write(tags_list)

    if tags_list is None or len(tags_list) == 0:
        return

    if append:
        if os.path.isfile(m3u8_file_name):
            old_tags = read_tags_from_m3u8(m3u8_file_name, False)
            old_tags.extend(tags_list)
            tags_list = old_tags

    # collect all file_names
    files = []
    for tags in tags_list:
        if 'SPOTY_FILE_NAME' in tags and tags['SPOTY_FILE_NAME'] != "":
            files.append(tags['SPOTY_FILE_NAME'] + "\n")

    os.makedirs(os.path.dirname(m3u8_file_name), exist_ok=True)
    with open(m3u8_file_name, 'w', encoding='utf-8-sig', newline='') as file:
        file.writelines(files)


def read_tags_from_m3u8s(m3u8_file_names, add_spoty_tags=True, clean_tags=True):
    all_tags_lists = []
    for m3u8_file_name in m3u8_file_names:
        tags_list = read_tags_from_m3u8(m3u8_file_name, add_spoty_tags, clean_tags)
        all_tags_lists.extend(tags_list)

    return all_tags_lists


def read_tags_from_m3u8(m3u8_file_name, add_spoty_tags=True, clean_tags=True):
    m3u8_file_name = os.path.abspath(m3u8_file_name)

    with open(m3u8_file_name, newline='', encoding='utf-8-sig') as file:
        files_list = file.readlines()
        for i, f in enumerate(files_list):
            files_list[i] = f.rstrip("\n").strip()

    tags_list = spoty.audio_files.read_audio_files_tags(files_list, add_spoty_tags, clean_tags)

    if add_spoty_tags:
        for i, tags in enumerate(tags_list):
            tags['SPOTY_PLAYLIST_NAME'] = os.path.splitext(os.path.basename(os.path.normpath(m3u8_file_name)))[0]
            tags['SPOTY_PLAYLIST_INDEX'] = str(i + 1)

    return tags_list



from spoty import settings
from spoty import log
import spoty.deezer_auth
import spoty.deezer_auth_manualy
import spoty.utils
import spoty.deezer_api
import click
import re
import time
import os
from datetime import datetime


@click.group()
def deezer():
    r"""Deezer specific commands."""
    pass


@deezer.group()
def auth():
    r"""Deezer authorization."""
    pass


@auth.command("get-token")
def auth_get_token():
    r"""
    Get deezer access token automatically using a local server.

    When the token is received, it will be written to the configuration file.
    """
    spoty.deezer_auth.get_token()
    pass


@auth.command("get-token-manualy")
def auth_get_token():
    r"""Get deezer access token manually.

    When starting the command, further actions will be described.
    """
    spoty.deezer_auth_manualy.get_token()
    pass


@deezer.group()
def playlist():
    r"""Playlists management."""
    pass


@playlist.command("list")
@click.option('--filter-names',
              help='List only playlists whose names matches this regex filter')
@click.option('--user-id',  help='Get playlists of this user')
def playlist_list(filter_names, user_id):
    r"""
    List of all playlists.

    Examples:

        spoty deezer playlist list

        spoty deezer playlist list --user-id 4717400682
    """
    if user_id == None:
        playlists = spoty.deezer_api.get_list_of_user_playlists()
    else:
        playlists = spoty.deezer_api.get_list_of_user_playlists(user_id)

    if len(playlists) == 0:
        exit()

    if filter_names is not None:
        playlists = list(filter(lambda pl: re.findall(filter_names, pl['title']), playlists))
        click.echo(f'{len(playlists)} playlists matches the filter')

    if len(playlists) == 0:
        exit()

    for playlist in playlists:
        click.echo(f'{playlist["id"]} "{playlist["title"]}"')

    click.echo(f'Total playlists: {len(playlists)}')


@playlist.command("create")
@click.argument("name", type=str)
def playlist_create(name):
    r"""
    Create a new empty playlist with the specified name.

    Examples:

        spoty deezer playlist create "My awesome playlist"
    """
    id = spoty.deezer_api.create_playlist(name)
    click.echo(f'New playlist created (id: {id}, name: "{name}")')


@playlist.command("delete")
@click.argument("playlist_id",  nargs=-1)
@click.option('--confirm', '-y',  is_flag=True, 
              help='Do not ask for export confirmation')
def playlist_delete(playlist_id, confirm):
    r"""
    Delete playlists with the specified id.

    PLAYLIST_IDS - list of playlist IDs or URIs. You can specify one ID or many IDs separated by a space.

    Examples:

        spoty deezer playlist delete 9457847341

        spoty deezer playlist delete 9457847341 --confirm

        spoty deezer playlist delete 9457847341 9712468342

        spoty deezer playlist delete https://www.deezer.com/ru/playlist/9457847341
    """
    deleted_playlists = spoty.deezer_api.delete_playlists(playlist_id, confirm)
    click.echo(f'{len(deleted_playlists)} playlist deleted')


@playlist.command("delete-all")
@click.option('--confirm', '-y',  is_flag=True, 
              help='Do not ask for export confirmation')
def playlist_delete_all(confirm):
    r"""
    Delete all playlists in the library.

    Examples:

        spoty deezer playlist delete-all

        spoty deezer playlist delete-all --confirm
    """
    deleted_playlists = spoty.deezer_api.delete_all_playlists(confirm)
    click.echo(f'{len(deleted_playlists)} playlist deleted')


# @playlist.command("copy")
# @click.argument("playlist_ids", type=str, nargs=-1)
# def playlist_copy(playlist_ids):
#     r"""
#     Create copies of playlists.
#     it could be your playlists or created by another user.
#     The exact same playlists will be created in your library.
#
#     PLAYLIST_IDS - list of playlist IDs or URIs. You can specify one ID or many IDs separated by a space.
#
#
#
#     Examples:
#
#         spoty deezer playlist copy 9457847341
#
#         spoty deezer playlist copy 9457847341 9712468342
#
#         spoty deezer playlist copy https://www.deezer.com/ru/playlist/9457847341
#
#     """
#
# playlists = []
# tracks = []
# with click.progressbar(playlist_ids, label='Copying playlists') as bar:
#     for playlist_id in bar:
#         new_playlist_id, tracks_added = spoty.deezer_api.copy_playlist(playlist_id)
#         playlists.extend(new_playlist_id)
#         tracks.extend(tracks_added)
#
# click.echo(f'{len(playlists)} playlists with {len(tracks)} tracks copied.')


@playlist.command("add-tracks")
@click.argument("playlist_id", )
@click.argument("track_ids",  nargs=-1)
@click.option('--allow-duplicates', '-d',  is_flag=True, 
              help='Add tracks that are already in the playlist.')
def playlist_add_tracks(playlist_id, track_ids, allow_duplicates):
    r"""
    Add tracks to playlist.

    PLAYLIST_ID - playlist ID or URI.

    TRACK_IDS - list of track IDs or URIs. You can specify one ID or many IDs separated by a space.

    Examples:

        spoty deezer playlist add-tracks 9457847341 74359352

        spoty deezer playlist add-tracks 9457847341 74359352 14456300

        spoty deezer playlist add-tracks https://www.deezer.com/ru/playlist/9457847341 https://www.deezer.com/ru/track/74359352

    """
    tracks_added, import_duplicates, already_exist \
        = spoty.deezer_api.add_tracks_to_playlist_by_ids(playlist_id, track_ids, allow_duplicates)
    click.echo(f'{len(tracks_added)} tracks added to playlist {playlist_id}')


@playlist.command("read")
@click.argument("playlist_ids",  nargs=-1)
def playlist_read(playlist_ids):
    r"""
    Read playlists and print track ids.
    it could be your playlists or created by another user.

    PLAYLIST_IDS - list of playlist IDs or URIs. You can specify one ID or many IDs separated by a space.

    Examples:

        spoty deezer playlist read 9457847341

        spoty deezer playlist read 9457847341 9712468342

        spoty deezer playlist read https://www.deezer.com/ru/playlist/9457847341

    """

    playlist_ids = spoty.utils.tuple_to_list(playlist_ids)
    playlists_dict = spoty.deezer_api.get_playlists_with_full_list_of_tracks(playlist_ids)

    for playlist_id, tracks_list in playlists_dict.items():
        click.echo(f'Tracks in playlist {playlist_id}:')
        for track in tracks_list:
            title = spoty.deezer_api.get_track_artist_and_title(track)
            click.echo(f'{track["SNG_ID"]} "{title}"')

        click.echo(f'Total tracks: {len(tracks_list)}')


@deezer.group()
def track():
    r"""Track management."""
    pass

@track.command("id")
@click.argument("id")
def track_find_by_id(id):
    r"""
    Find track by ID.
    """
    track = spoty.deezer_api.find_track_by_id(id)
    if track == None:
        click.echo("Not found")
        return
    tags = spoty.deezer_api.read_tags_from_deezer_track(track)
    spoty.utils.print_main_tags(tags)



from spoty.commands.first_list_commands import \
    count_command, \
    delete_command, \
    copy_command, \
    move_command, \
    export_command, \
    import_deezer_command, \
    import_spotify_command, \
    sync_spotify_command, \
    print_command, \
    find_duplicates_command, \
    create_m3u8_command, \
    filter_group,\
    find_deezer_group, \
    find_spotify_group
from spoty.commands import \
    get_second_group
from spoty.utils import SpotyContext
import spoty.spotify_api
import spoty.deezer_api
import spoty.audio_files
import spoty.csv_playlist
import spoty.m3u8_playlist
import spoty.utils
import click


@click.group("get")
@click.option('--spotify-playlist', '--sp', multiple=True,
              help='Get tracks from Spotify playlist URI or ID.')
@click.option('--spotify-entire-library', '--s', multiple=True,
              help='Get all tracks from Spotify library (by user URI or ID). To request a list for the current authorized user, use "me" as ID.')
@click.option('--spotify-entire-library-regex', '--sr', nargs=2, multiple=True,
              help='Works the same as --spotify-entire-library, but you can specify regex filter which will be applied to playlists names. This way you can query any playlists by names. First param - user ID, second param - regex query.')
@click.option('--deezer-playlist', '--dp', multiple=True,
              help='Get tracks from Deezer playlist URI or ID.')
@click.option('--deezer-entire-library', '--d', multiple=True,
              help='Get all tracks from Deezer library (by user URI or ID). To request a list for the current authorized user, use "me" as ID.')
@click.option('--deezer-entire-library-regex', '--dr', nargs=2, multiple=True,
              help='Works the same as --deezer-entire-library, but you can specify regex filter which will be applied to playlists names. This way you can query any playlists by names. First param - user ID, second param - regex query.')
@click.option('--audio', '--a', multiple=True,
              help='Get audio files located at the specified local path. You can specify the audio file name as well.')
@click.option('--csv', '--c', multiple=True,
              help='Get tracks from csv playlists located at the specified local path. You can specify the scv file name as well.')
@click.option('--m3u8', '--m', multiple=True,
              help='Get tracks from m3u8 playlists located at the specified local path. You can specify the m3u8 file name as well.')
@click.option('--no-recursive', '-r', is_flag=True,
              help='Do not search in subdirectories from the specified path.')
@click.pass_context
def get_tracks(
        ctx,
        spotify_playlist,
        spotify_entire_library,
        spotify_entire_library_regex,
        deezer_playlist,
        deezer_entire_library,
        deezer_entire_library_regex,
        audio,
        csv,
        m3u8,
        no_recursive,
):
    """
Get tracks from sources for further actions (see next commands).
    """

    if ctx.obj is None:
        ctx.obj = SpotyContext()

    ctx.obj.summary.append("Collecting:")

    get_tracks_wrapper(
        ctx.obj,
        spotify_playlist,
        spotify_entire_library,
        spotify_entire_library_regex,
        deezer_playlist,
        deezer_entire_library,
        deezer_entire_library_regex,
        audio,
        csv,
        m3u8,
        no_recursive,
    )


def get_tracks_wrapper(
        context: SpotyContext,
        spotify_playlist,
        spotify_entire_library,
        spotify_entire_library_regex,
        deezer_playlist,
        deezer_entire_library,
        deezer_entire_library_regex,
        audio,
        csv,
        m3u8,
        no_recursive,
):
    all_tags_list = []

    # get m3u8

    m3u8_files = []
    tags_list_from_m3u8 = []

    if len(m3u8) > 0:
        m3u8_paths = []
        for path in m3u8:
            if spoty.m3u8_playlist.is_m3u8(path):
                if spoty.utils.is_valid_file(path):
                    m3u8_files.append(path)
            elif spoty.utils.is_valid_path(path):
                m3u8_paths.append(path)
            else:
                click.echo(f'Cant find path or file: "{path}"', err=True)
                exit()

        file_names = spoty.m3u8_playlist.find_m3u8s_in_paths(m3u8_paths, not no_recursive)
        m3u8_files.extend(file_names)

        tags_list = spoty.m3u8_playlist.read_tags_from_m3u8s(m3u8_files)
        tags_list_from_m3u8.extend(tags_list)
        all_tags_list.extend(tags_list)


    # get csv

    csv_files = []
    tags_list_from_csv = []

    if len(csv) > 0:
        csv_paths = []
        for path in csv:
            if spoty.csv_playlist.is_csv(path):
                if spoty.utils.is_valid_file(path):
                    csv_files.append(path)
            elif spoty.utils.is_valid_path(path):
                csv_paths.append(path)
            else:
                click.echo(f'Cant find path or file: "{path}"', err=True)
                exit()

        file_names = spoty.csv_playlist.find_csvs_in_paths(csv_paths, not no_recursive)
        csv_files.extend(file_names)

        tags_list = spoty.csv_playlist.read_tags_from_csvs(csv_files)
        tags_list_from_csv.extend(tags_list)
        all_tags_list.extend(tags_list)

    # get audio

    audio_files = []
    tags_list_from_audio = []

    if len(audio) > 0:
        audio_paths = []
        for path in audio:
            if spoty.audio_files.is_audio_file(path):
                if spoty.utils.is_valid_file(path):
                    audio_files.append(path)
            elif spoty.utils.is_valid_path(path):
                audio_paths.append(path)
            else:
                click.echo(f'Cant find path or file: "{path}"', err=True)
                exit()

        file_names = spoty.audio_files.find_audio_files_in_paths(audio_paths, not no_recursive)
        audio_files.extend(file_names)

        tags_list = spoty.audio_files.read_audio_files_tags(audio_files)
        tags_list = spoty.utils.add_playlist_index_from_playlist_names(tags_list)
        tags_list_from_audio.extend(tags_list)
        all_tags_list.extend(tags_list)

    # get spotify

    spotify_playlists = []
    tags_list_from_spotify = []

    if len(spotify_playlist) > 0:
        pl = spoty.utils.tuple_to_list(spotify_playlist)
        tracks, tags_list, playlists = spoty.spotify_api.get_tracks_from_playlists(pl)
        spotify_playlists.extend(playlists)
        tags_list_from_spotify.extend(tags_list)
        all_tags_list.extend(tags_list)

    if len(spotify_entire_library) > 0:
        for user_id in spotify_entire_library:
            tracks, tags_list, playlists = spoty.spotify_api.get_tracks_of_spotify_user(user_id)
            spotify_playlists.extend(playlists)
            tags_list_from_spotify.extend(tags_list)
            all_tags_list.extend(tags_list)

    if len(spotify_entire_library_regex) > 0:
        for user_and_reg in spotify_entire_library_regex:
            tracks, tags_list, playlists = spoty.spotify_api.get_tracks_of_spotify_user(user_and_reg[0],
                                                                                        user_and_reg[1])
            spotify_playlists.extend(playlists)
            tags_list_from_spotify.extend(tags_list)
            all_tags_list.extend(tags_list)

    # get deezer

    deezer_playlists = []
    tags_list_from_deezer = []

    if len(deezer_playlist) > 0:
        pl = spoty.utils.tuple_to_list(deezer_playlist)
        tracks, tags_list, playlists = spoty.deezer_api.get_tracks_from_playlists(pl)
        # tags_list = spoty.deezer_api.add_track_release_dates(tags_list)
        deezer_playlists.extend(playlists)
        tags_list_from_deezer.extend(tags_list)
        all_tags_list.extend(tags_list)

    if len(deezer_entire_library) > 0:
        for user_id in deezer_entire_library:
            tracks, tags_list, playlists = spoty.deezer_api.get_tracks_of_deezer_user(user_id)
            deezer_playlists.extend(playlists)
            tags_list_from_deezer.extend(tags_list)
            all_tags_list.extend(tags_list)

    if len(deezer_entire_library_regex) > 0:
        for user_and_reg in deezer_entire_library_regex:
            tracks, tags_list, playlists = spoty.deezer_api.get_tracks_of_deezer_user(user_and_reg[0], user_and_reg[1])
            deezer_playlists.extend(playlists)
            tags_list_from_deezer.extend(tags_list)
            all_tags_list.extend(tags_list)

    # make summary

    summary = []

    if len(tags_list_from_spotify) > 0:
        context.summary.append(
            f'  {len(tags_list_from_spotify)} tracks found in {len(spotify_playlists)} Spotify playlists.')
    if len(tags_list_from_deezer) > 0:
        context.summary.append(
            f'  {len(tags_list_from_deezer)} tracks found in {len(deezer_playlists)} Deezer playlists.')
    if len(tags_list_from_audio) > 0:
        context.summary.append(f'  {len(tags_list_from_audio)} audio files found in local path.')
    if len(tags_list_from_csv) > 0:
        context.summary.append(f'  {len(tags_list_from_csv)} tracks found in {len(csv_files)} csv playlists.')
    if len(tags_list_from_m3u8) > 0:
        context.summary.append(f'  {len(tags_list_from_m3u8)} tracks found in {len(m3u8_files)} m3u8 playlists.')

    if not (len(tags_list_from_spotify) == len(all_tags_list) or len(tags_list_from_spotify) == 0) \
            or not (len(tags_list_from_deezer) == len(all_tags_list) or len(tags_list_from_deezer) == 0) \
            or not (len(tags_list_from_audio) == len(all_tags_list) or len(tags_list_from_audio) == 0) \
            or not (len(tags_list_from_csv) == len(all_tags_list) or len(tags_list_from_csv) == 0)\
            or not (len(tags_list_from_m3u8) == len(all_tags_list) or len(tags_list_from_m3u8) == 0):
        context.summary.append(f'  {len(all_tags_list)} total tracks collected.')

    if len(all_tags_list) == 0:
        context.summary.append(f'  {len(all_tags_list)} total tracks collected.')

    # make context

    context.tags_lists.append(all_tags_list)


get_tracks.add_command(filter_group.filter_tracks)

get_tracks.add_command(count_command.count_tracks)
get_tracks.add_command(print_command.print_tracks)
get_tracks.add_command(export_command.export_tracks)
get_tracks.add_command(create_m3u8_command.export_tracks)
get_tracks.add_command(import_spotify_command.import_spotify)
get_tracks.add_command(import_deezer_command.import_deezer)
get_tracks.add_command(sync_spotify_command.sync_spotify)
get_tracks.add_command(find_deezer_group.find_deezer)
get_tracks.add_command(find_spotify_group.find_spotify)
get_tracks.add_command(get_second_group.get_second)
get_tracks.add_command(delete_command.delete_tracks)
get_tracks.add_command(move_command.move_tracks)
get_tracks.add_command(copy_command.copy_tracks)
get_tracks.add_command(find_duplicates_command.find_duplicates)
